from optimagic.decorators import deprecated
from optimagic.utilities import (
    calculate_trustregion_initial_radius as _calculate_trustregion_initial_radius,
)
from optimagic.utilities import (
    chol_params_to_lower_triangular_matrix as _chol_params_to_lower_triangular_matrix,
)
from optimagic.utilities import cov_matrix_to_params as _cov_matrix_to_params
from optimagic.utilities import (
    cov_matrix_to_sdcorr_params as _cov_matrix_to_sdcorr_params,
)
from optimagic.utilities import cov_params_to_matrix as _cov_params_to_matrix
from optimagic.utilities import cov_to_sds_and_corr as _cov_to_sds_and_corr
from optimagic.utilities import (
    dimension_to_number_of_triangular_elements as _dimension_to_number_of_triangular_elements,  # noqa: E501
)
from optimagic.utilities import get_rng as _get_rng
from optimagic.utilities import hash_array as _hash_array
from optimagic.utilities import isscalar as _isscalar
from optimagic.utilities import (
    number_of_triangular_elements_to_dimension as _number_of_triangular_elements_to_dimension,  # noqa: E501
)
from optimagic.utilities import propose_alternatives as _propose_alternatives
from optimagic.utilities import read_pickle as _read_pickle
from optimagic.utilities import robust_cholesky as _robust_cholesky
from optimagic.utilities import robust_inverse as _robust_inverse
from optimagic.utilities import sdcorr_params_to_matrix as _sdcorr_params_to_matrix
from optimagic.utilities import (
    sdcorr_params_to_sds_and_corr as _sdcorr_params_to_sds_and_corr,
)
from optimagic.utilities import sds_and_corr_to_cov as _sds_and_corr_to_cov
from optimagic.utilities import to_pickle as _to_pickle

MSG = (
    "estimagic.utilities.{name} has been deprecated in version 0.5.0. Use optimagic."
    "utilities.{name} instead. This function will be removed in version 0.6.0."
)


chol_params_to_lower_triangular_matrix = deprecated(
    _chol_params_to_lower_triangular_matrix,
    MSG.format(name="chol_params_to_lower_triangular_matrix"),
)
cov_params_to_matrix = deprecated(
    _cov_params_to_matrix, MSG.format(name="cov_params_to_matrix")
)
cov_matrix_to_params = deprecated(
    _cov_matrix_to_params, MSG.format(name="cov_matrix_to_params")
)
sdcorr_params_to_sds_and_corr = deprecated(
    _sdcorr_params_to_sds_and_corr, MSG.format(name="sdcorr_params_to_sds_and_corr")
)
sds_and_corr_to_cov = deprecated(
    _sds_and_corr_to_cov, MSG.format(name="sds_and_corr_to_cov")
)
cov_to_sds_and_corr = deprecated(
    _cov_to_sds_and_corr, MSG.format(name="cov_to_sds_and_corr")
)
sdcorr_params_to_matrix = deprecated(
    _sdcorr_params_to_matrix, MSG.format(name="sdcorr_params_to_matrix")
)
cov_matrix_to_sdcorr_params = deprecated(
    _cov_matrix_to_sdcorr_params, MSG.format(name="cov_matrix_to_sdcorr_params")
)
number_of_triangular_elements_to_dimension = deprecated(
    _number_of_triangular_elements_to_dimension,
    MSG.format(name="number_of_triangular_elements_to_dimension"),
)
dimension_to_number_of_triangular_elements = deprecated(
    _dimension_to_number_of_triangular_elements,
    MSG.format(name="dimension_to_number_of_triangular_elements"),
)
propose_alternatives = deprecated(
    _propose_alternatives, MSG.format(name="propose_alternatives")
)
robust_cholesky = deprecated(_robust_cholesky, MSG.format(name="robust_cholesky"))
robust_inverse = deprecated(_robust_inverse, MSG.format(name="robust_inverse"))
hash_array = deprecated(_hash_array, MSG.format(name="hash_array"))
calculate_trustregion_initial_radius = deprecated(
    _calculate_trustregion_initial_radius,
    MSG.format(name="calculate_trustregion_initial_radius"),
)
to_pickle = deprecated(_to_pickle, MSG.format(name="to_pickle"))
read_pickle = deprecated(_read_pickle, MSG.format(name="read_pickle"))
isscalar = deprecated(_isscalar, MSG.format(name="isscalar"))
get_rng = deprecated(_get_rng, MSG.format(name="get_rng"))

__all__ = [
    "chol_params_to_lower_triangular_matrix",
    "cov_params_to_matrix",
    "cov_matrix_to_params",
    "sdcorr_params_to_sds_and_corr",
    "sds_and_corr_to_cov",
    "cov_to_sds_and_corr",
    "sdcorr_params_to_matrix",
    "cov_matrix_to_sdcorr_params",
    "number_of_triangular_elements_to_dimension",
    "dimension_to_number_of_triangular_elements",
    "propose_alternatives",
    "robust_cholesky",
    "robust_inverse",
    "hash_array",
    "calculate_trustregion_initial_radius",
    "to_pickle",
    "read_pickle",
    "isscalar",
    "get_rng",
]



from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar

from optimagic.optimization.algorithm import AlgoInfo
from optimagic.typing import AggregationLevel

P = ParamSpec("P")


ScalarFuncT = TypeVar("ScalarFuncT", bound=Callable[..., Any])
VectorFuncT = TypeVar("VectorFuncT", bound=Callable[..., Any])


def scalar(func: ScalarFuncT) -> ScalarFuncT:
    """Mark a function as a scalar function."""
    wrapper = func
    try:
        wrapper._problem_type = AggregationLevel.SCALAR  # type: ignore
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:

        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            return func(*args, **kwargs)

        wrapper._problem_type = AggregationLevel.SCALAR  # type: ignore
    return wrapper


def least_squares(func: VectorFuncT) -> VectorFuncT:
    """Mark a function as a least squares function."""
    wrapper = func
    try:
        wrapper._problem_type = AggregationLevel.LEAST_SQUARES  # type: ignore
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:

        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            return func(*args, **kwargs)

        wrapper._problem_type = AggregationLevel.LEAST_SQUARES  # type: ignore
    return wrapper


def likelihood(func: VectorFuncT) -> VectorFuncT:
    """Mark a function as a likelihood function."""
    wrapper = func
    try:
        wrapper._problem_type = AggregationLevel.LIKELIHOOD  # type: ignore
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:

        @wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            return func(*args, **kwargs)

        wrapper._problem_type = AggregationLevel.LIKELIHOOD  # type: ignore
    return wrapper


# TODO: I get an error when adding bound=Algorithm to AlgorithmSubclass. Why?
AlgorithmSubclass = TypeVar("AlgorithmSubclass")


def minimizer(
    name: str,
    solver_type: AggregationLevel,
    is_available: bool,
    is_global: bool,
    needs_jac: bool,
    needs_hess: bool,
    supports_parallelism: bool,
    supports_bounds: bool,
    supports_linear_constraints: bool,
    supports_nonlinear_constraints: bool,
    disable_history: bool = False,
) -> Callable[[AlgorithmSubclass], AlgorithmSubclass]:
    """Mark an algorithm as a optimagic minimizer and add AlgoInfo.

    Args:
        name: The name of the algorithm as a string. Used in error messages, warnings
            and the OptimizeResult.
        solver_type: The type of optimization problem the algorithm solves. Used to
            distinguish between scalar, least-squares and likelihood optimizers. Can
            take the values AggregationLevel.SCALAR, AggregationLevel.LEAST_SQUARES and
            AggregationLevel.LIKELIHOOD.
        is_available: Whether the algorithm is installed.
        is_global: Whether the algorithm is a global optimizer.
        needs_jac: Whether the algorithm needs some kind of first derivative. This needs
            to be True if the algorithm uses `jac` or `fun_and_jac`.
        needs_hess: Whether the algorithm needs some kind of second derivative. This
            is not yet implemented and will be False for all currently wrapped
            algorithms.
        supports_parallelism: Whether the algorithm supports parallelism. This needs to
            be True if the algorithm previously took `n_cores` and/or `batch_evaluator`
            as arguments.
        supports_bounds: Whether the algorithm supports bounds. This needs to be True
            if the algorithm previously took `lower_bounds` and/or `upper_bounds` as
            arguments.
        supports_linear_constraints: Whether the algorithm supports linear constraints.
            This is not yet implemented and will be False for all currently wrapped
            algorithms.
        supports_nonlinear_constraints: Whether the algorithm supports nonlinear
            constraints. This needs to be True if the algorithm previously took
            `nonlinear_constraints` as an argument.
        disable_history: Whether the algorithm should disable history collection.

    """

    def decorator(cls: AlgorithmSubclass) -> AlgorithmSubclass:
        algo_info = AlgoInfo(
            name=name,
            solver_type=solver_type,
            is_available=is_available,
            is_global=is_global,
            needs_jac=needs_jac,
            needs_hess=needs_hess,
            supports_parallelism=supports_parallelism,
            supports_bounds=supports_bounds,
            supports_linear_constraints=supports_linear_constraints,
            supports_nonlinear_constraints=supports_nonlinear_constraints,
            disable_history=disable_history,
        )
        cls.__algo_info__ = algo_info  # type: ignore
        return cls

    return decorator
